use std::{error::Error, str::FromStr, sync::Arc};

use config::Config;
use futures::future::join_all;
use serde::Deserialize;
use solana_client::nonblocking::rpc_client::RpcClient;
use solana_sdk::{
    commitment_config::CommitmentConfig, native_token::LAMPORTS_PER_SOL, pubkey::Pubkey,
};

#[derive(Deserialize)]
struct Cfg {
    accounts: Vec<String>,
    solana_url: String,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn Error>> {
    let config = {
        let config_parser = Config::builder()
            .add_source(config::File::with_name("config.yaml"))
            .build()?;

        config_parser.try_deserialize::<Cfg>()?
    };
    let client = Arc::new(RpcClient::new_with_commitment(
        config.solana_url.clone(),
        CommitmentConfig::confirmed(),
    ));

    let mut fetchers = vec![];
    for account in &config.accounts {
        let client = client.clone();
        fetchers.push(async move {
            let pubkey = Pubkey::from_str(&account)?;
            let balance = client.get_balance(&pubkey).await?;
            Ok::<_, Box<dyn Error>>(balance as f64 / LAMPORTS_PER_SOL as f64)
        });
    }

    let balances = join_all(fetchers)
        .await
        .into_iter()
        .collect::<Result<Vec<_>, _>>()?;

    for (acc, bal) in config.accounts.iter().zip(balances.iter()) {
        println!("{acc}: {bal} SOL");
    }

    Ok(())
}
